@@ -1,7 +1,8 @@
 """PCK command parsers and generators."""
 
+from __future__ import annotations
+
 import re
-from typing import Optional
 from collections.abc import Sequence
 
 from pypck import lcn_defs
@@ -42,6 +43,9 @@ class PckParser:
     # Pattern to parse error messages.
     PATTERN_COMMAND_ERROR = re.compile(r"\((?P<message>.+)\?\)")
 
+    # Pattern to parse ping messages.
+    PATTERN_PING = re.compile(r"\^ping(?P<count>\d*)-")
+
     # Pattern to parse positive acknowledges.
     PATTERN_ACK_POS = re.compile(r"-M(?P<seg_id>\d{3})(?P<mod_id>\d{3})!")
 
@@ -150,7 +154,7 @@ class PckParser:
 
     # Pattern to parse send command to host messages.
     PATTERN_SEND_COMMAND_HOST = re.compile(
-        r"\+M004(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SKH"
+        r"(\$M|\+M004)(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SKH"
         r"(?P<p1>\d{3})(?P<p2>\d{3})"
         r"(?:(?P<p3>\d{3})(?P<p4>\d{3})(?P<p5>\d{3})(?P<p6>\d{3}))?"
         r"(?:(?P<p7>\d{3})(?P<p8>\d{3})(?P<p9>\d{3})(?P<p10>\d{3})"
@@ -159,7 +163,7 @@ class PckParser:
 
     # Pattern to parse send key to host messages.
     PATTERN_SEND_KEYS_HOST = re.compile(
-        r"\+M004(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.STH"
+        r"(\$M|\+M004)(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.STH"
         r"(?P<actions>\d{3})(?P<keys>\d{3})"
     )
 
@@ -571,7 +575,7 @@ class PckGenerator:
     @staticmethod
     def control_motors_outputs(
         state: lcn_defs.MotorStateModifier,
-        reverse_time: Optional[lcn_defs.MotorReverseTime] = None,
+        reverse_time: lcn_defs.MotorReverseTime | None = None,
     ) -> str:
         """Generate a command to control a motor via output ports 1+2.
 
@@ -585,24 +589,24 @@ class PckGenerator:
         if state == lcn_defs.MotorStateModifier.UP:
             if reverse_time in [None, lcn_defs.MotorReverseTime.RT70]:
                 params = (0x01, 0xE4, 0x00)
+                ret = f"X2{params[0]:03d}{params[1]:03d}{params[2]:03d}"
             elif reverse_time == lcn_defs.MotorReverseTime.RT600:
-                params = (0x04, 0xC8, 0x08)
+                ret = PckGenerator.dim_output(0, 100, 8)
             elif reverse_time == lcn_defs.MotorReverseTime.RT1200:
-                params = (0x04, 0xC8, 0x0B)
+                ret = PckGenerator.dim_output(0, 100, 11)
             else:
                 raise ValueError("Wrong MotorReverseTime.")
-            ret = f"X2{params[0]:03d}{params[1]:03d}{params[2]:03d}"
 
         elif state == lcn_defs.MotorStateModifier.DOWN:
             if reverse_time in [None, lcn_defs.MotorReverseTime.RT70]:
                 params = (0x01, 0x00, 0xE4)
+                ret = f"X2{params[0]:03d}{params[1]:03d}{params[2]:03d}"
             elif reverse_time == lcn_defs.MotorReverseTime.RT600:
-                params = (0x05, 0xC8, 0x08)
+                ret = PckGenerator.dim_output(1, 100, 8)
             elif reverse_time == lcn_defs.MotorReverseTime.RT1200:
-                params = (0x05, 0xC8, 0x0B)
+                ret = PckGenerator.dim_output(1, 100, 11)
             else:
                 raise ValueError("Wrong MotorReverseTime.")
-            ret = f"X2{params[0]:03d}{params[1]:03d}{params[2]:03d}"
 
         elif state == lcn_defs.MotorStateModifier.STOP:
             ret = "AY000000"
@@ -1025,16 +1029,29 @@ class PckGenerator:
         return f"GTDT{row_id + 1}{part_id + 1}".encode() + part
 
     @staticmethod
-    def lock_regulator(reg_id: int, state: bool) -> str:
+    def lock_regulator(
+        reg_id: int,
+        state: bool,
+        software_serial: int,
+        target_value: float = -1,
+    ) -> str:
         """Generate a command to lock a regulator.
 
         :param    int    reg_id:    Regulator id 0..1
         :param    bool   state:     The lock state
+        :param    int    software_serial: The expected firmware version of all
+                                           receiving modules.
+        :param    foat    target_value: The target value in percent (use -1 to ignore)
         :return:  The PCK command (without address header) as text
         :rtype:   str
         """
         if (reg_id < 0) or (reg_id > 1):
             raise ValueError("Wrong reg_id.")
+        if ((target_value < 0) or (target_value > 100)) and (target_value != -1):
+            raise ValueError("Wrong target_value.")
+        if (target_value != -1) and (software_serial >= 0x120301) and state:
+            reg_byte = reg_id * 0x40 + 0x07
+            return f"X2{0x1E:03d}{reg_byte:03d}{int(2*target_value):03d}"
         return f"RE{'A' if reg_id == 0 else 'B'}X{'S' if state else 'A'}"
 
     @staticmethod
@@ -1077,7 +1094,7 @@ class PckGenerator:
     def activate_scene_output(
         scene_id: int,
         output_ports: Sequence[lcn_defs.OutputPort] = (),
-        ramp: Optional[int] = None,
+        ramp: int | None = None,
     ) -> str:
         """Activate the stored output states for the given scene.
 
@@ -1100,7 +1117,7 @@ class PckGenerator:
     def store_scene_output(
         scene_id: int,
         output_ports: Sequence[lcn_defs.OutputPort] = (),
-        ramp: Optional[int] = None,
+        ramp: int | None = None,
     ) -> str:
         """Store the current output states in the given scene.
 
@@ -1123,7 +1140,7 @@ class PckGenerator:
     def _activate_or_store_scene_output(
         scene_id: int,
         output_ports: Sequence[lcn_defs.OutputPort] = (),
-        ramp: Optional[int] = None,
+        ramp: int | None = None,
         store: bool = False,
     ) -> str:
         if (scene_id < 0) or (scene_id > 9):
